@@ -1,5 +1,6 @@
 import base64
 import getpass
+import hashlib
 import itertools
 import platform
 import random
@@ -14,13 +15,13 @@ import websocket
 import zlib
 
 from . import callback as cb
-from . import constant
 
 from .exceptions import *
 from .statement import ExaStatement
 from .logger import ExaLogger
 from .formatter import ExaFormatter
 from .ext import ExaExtension
+from .http_transport import ExaSQLExportThread, ExaSQLImportThread, ExaHttpThread
 from .meta import ExaMetaData
 from .script_output import ExaScriptOutputProcess
 from .version import __version__
@@ -95,7 +96,7 @@ class ExaConnection(object):
         :param fetch_size_bytes: Maximum size of data message for single fetch request in bytes (Default: 5Mb)
         :param lower_ident: Automatically lowercase identifiers (table names, column names, etc.) returned from relevant functions (Default: False)
         :param quote_ident: Add double quotes and escape identifiers passed to relevant functions (export_*, import_*, ext.*, etc.) (Default: False)
-        :param json_lib: Supported values: rapidjson, ujson, json (Default: json)
+        :param json_lib: Supported values: rapidjson, ujson, orjson, json (Default: json)
         :param verbose_error: Display additional information when error occurs (Default: True)
         :param debug: Output debug information for client-server communication and connection attempts to STDERR
         :param debug_logdir: Store debug information into files in debug_logdir instead of outputting it to STDERR
@@ -160,7 +161,7 @@ class ExaConnection(object):
         self.attr = {}
         self.is_closed = False
 
-        self.ws_host = None
+        self.ws_ipaddr = None
         self.ws_port = None
         self.ws_req_count = 0
         self.ws_req_time = 0
@@ -197,7 +198,7 @@ class ExaConnection(object):
         Execute SQL query with UDF script, capture output
         Return ExaStatement object and list of Path-objects for script output log files
 
-        Exasol should be able to open connection to the host where current script is running
+        Exasol should be able to open connection to the machine where current script is running
         """
         stmt_output_dir = self._get_stmt_output_dir()
 
@@ -287,8 +288,6 @@ class ExaConnection(object):
         return self.import_from_callback(cb.import_from_pandas, src, table, callback_params, import_params)
 
     def export_to_callback(self, callback, dst, query_or_table, query_params=None, callback_params=None, export_params=None):
-        from .http_transport import ExaSQLExportThread, ExaHTTPProcess, HTTP_EXPORT
-
         if not callable(callback):
             raise ValueError('Callback argument is not callable')
 
@@ -298,150 +297,126 @@ class ExaConnection(object):
         if export_params is None:
             export_params = {}
 
-        if 'format' in export_params:
-            compression = False
-        else:
-            compression = self.options['compression']
-
         if query_params is not None:
             query_or_table = self.format.format(query_or_table, **query_params)
 
-        http_proc = ExaHTTPProcess(self.ws_host, self.ws_port, compression, self.options['encryption'], HTTP_EXPORT)
+        compression = False if ('format' in export_params) else self.options['compression']
+
+        http_thread = ExaHttpThread(self.ws_ipaddr, self.ws_port, compression, self.options['encryption'])
         sql_thread = ExaSQLExportThread(self, compression, query_or_table, export_params)
 
         try:
-            http_proc.start()
-
-            sql_thread.set_http_proc(http_proc)
-            sql_thread.set_exa_proxy_list(http_proc.get_proxy())
+            http_thread.start()
 
+            sql_thread.set_http_thread(http_thread)
             sql_thread.start()
 
-            result = callback(http_proc.read_pipe, dst, **callback_params)
-            http_proc.read_pipe.close()
+            result = callback(http_thread.read_pipe, dst, **callback_params)
+            http_thread.read_pipe.close()
 
-            http_proc.join_with_exc()
+            http_thread.join_with_exc()
             sql_thread.join_with_exc()
 
             return result
-        except Exception as e:
-            # Terminate HTTP Server if it is still running
-            http_proc.terminate()
-            http_proc.join()
 
-            # Try to join SQL thread, but no longer than 1 second
+        except (Exception, KeyboardInterrupt) as e:
+            http_thread.terminate_export()
+            http_thread.join()
+
             sql_thread.join(1)
 
-            # If SQL thread is still running somehow, abort query and join again
+            # Prevent infinite lock if SQL query is still running
             if sql_thread.is_alive():
                 self.abort_query()
                 sql_thread.join()
 
-            # Give higher priority to SQL thread exception
+            # Give SQL exception higher priority
             if sql_thread.exc:
                 raise sql_thread.exc
 
             raise e
 
     def import_from_callback(self, callback, src, table, callback_params=None, import_params=None):
-        from .http_transport import ExaSQLImportThread, ExaHTTPProcess, HTTP_IMPORT
-
         if callback_params is None:
             callback_params = {}
 
         if import_params is None:
             import_params = {}
 
-        if 'format' in import_params:
-            compression = False
-        else:
-            compression = self.options['compression']
+        compression = False if ('format' in import_params) else self.options['compression']
 
         if not callable(callback):
             raise ValueError('Callback argument is not callable')
 
-        http_proc = ExaHTTPProcess(self.ws_host, self.ws_port, compression, self.options['encryption'], HTTP_IMPORT)
+        http_thread = ExaHttpThread(self.ws_ipaddr, self.ws_port, compression, self.options['encryption'])
         sql_thread = ExaSQLImportThread(self, compression, table, import_params)
 
         try:
-            http_proc.start()
-
-            sql_thread.set_http_proc(http_proc)
-            sql_thread.set_exa_proxy_list(http_proc.get_proxy())
+            http_thread.start()
 
+            sql_thread.set_http_thread(http_thread)
             sql_thread.start()
 
-            result = callback(http_proc.write_pipe, src, **callback_params)
-            http_proc.write_pipe.close()
+            result = callback(http_thread.write_pipe, src, **callback_params)
+            http_thread.write_pipe.close()
 
-            http_proc.join_with_exc()
+            http_thread.join_with_exc()
             sql_thread.join_with_exc()
 
             return result
-        except Exception as e:
-            # Terminate HTTP Server if it is still running
-            http_proc.terminate()
-            http_proc.join()
 
-            # Try to join SQL thread, but no longer than 1 second
+        except (Exception, KeyboardInterrupt) as e:
+            http_thread.terminate_import()
+            http_thread.join()
+
             sql_thread.join(1)
 
-            # If SQL thread is still running somehow, abort query in the main thread and join SQL thread again
+            # Prevent infinite lock if SQL query is still running
             if sql_thread.is_alive():
                 self.abort_query()
                 sql_thread.join()
 
-            # Give higher priority to SQL thread exception
+            # Give SQL exception higher priority
             if sql_thread.exc:
                 raise sql_thread.exc
 
             raise e
 
-    def export_parallel(self, exa_proxy_list, query_or_table, query_params=None, export_params=None):
+    def export_parallel(self, exa_address_list, query_or_table, query_params=None, export_params=None):
         """
         Init HTTP transport in child processes first using pyexasol.http_transport()
-        Get proxy strings from each child process
-        Pass proxy strings to parent process and use it for export_parallel() call
+        Get internal Exasol address from each child process using .address
+        Pass address strings to parent process, combine into single list and use it for export_parallel() call
         """
-        from .http_transport import ExaSQLExportThread
-
         if export_params is None:
             export_params = {}
 
-        if 'format' in export_params:
-            compression = False
-        else:
-            compression = self.options['compression']
+        compression = False if ('format' in export_params) else self.options['compression']
 
         if query_params is not None:
             query_or_table = self.format.format(query_or_table, **query_params)
 
-        # There is no need to run separate thread here, all work is performed in child processes
+        # There is no need to actually run a separate thread here, all work is performed in separate processes
         # We simply reuse thread class to keep logic in one place
         sql_thread = ExaSQLExportThread(self, compression, query_or_table, export_params)
-        sql_thread.set_exa_proxy_list(exa_proxy_list)
+        sql_thread.set_exa_address_list(exa_address_list)
         sql_thread.run_sql()
 
-    def import_parallel(self, exa_proxy_list, table, import_params=None):
+    def import_parallel(self, exa_address_list, table, import_params=None):
         """
         Init HTTP transport in child processes first using pyexasol.http_transport()
-        Get proxy strings from each child process
-        Pass proxy strings to parent process and use it for import_parallel() call
+        Get internal Exasol address from each child process using .address
+        Pass address strings to parent process, combine into single list and use it for import_parallel() call
         """
-        from .http_transport import ExaSQLImportThread
-
         if import_params is None:
             import_params = {}
 
-        if 'format' in import_params:
-            compression = False
-        else:
-            compression = self.options['compression']
+        compression = False if ('format' in import_params) else self.options['compression']
 
-        # There is no need to run separate thread here, all work is performed in child processes
+        # There is no need to actually run a separate thread here, all work is performed in separate processes
         # We simply reuse thread class to keep logic in one place
         sql_thread = ExaSQLImportThread(self, compression, table, import_params)
-        sql_thread.set_exa_proxy_list(exa_proxy_list)
+        sql_thread.set_exa_address_list(exa_address_list)
         sql_thread.run_sql()
 
     def session_id(self):
@@ -507,7 +482,7 @@ class ExaConnection(object):
     def get_nodes(self, pool_size=None):
         """
         Return list of dictionaries describing active Exasol nodes
-        Format: {'host': <ip_address>, 'port': <port>, 'idx': <incremental index of returned node>}
+        Format: {'ipaddr': <ip_address>, 'port': <port>, 'idx': <incremental index of returned node>}
 
         If pool_size is bigger than number of nodes, list will wrap around and nodes will repeat with different 'idx'
         If pool_size is omitted, return every active node once
@@ -517,13 +492,14 @@ class ExaConnection(object):
         """
         ret = self.req({
             'command': 'getHosts',
-            'hostIp': self.ws_host,
+            'hostIp': self.ws_ipaddr,
         })
 
         if pool_size is None:
             pool_size = ret['responseData']['numNodes']
 
-        return [{'host': ip_address, 'port': self.ws_port, 'idx': idx} for idx, ip_address
+        # Key 'host' is deprecated and remains only for backwards compatibility, please use `ipaddr` instead
+        return [{'host': ipaddr, 'ipaddr': ipaddr, 'port': self.ws_port, 'idx': idx} for idx, ipaddr
                 in enumerate(itertools.islice(itertools.cycle(ret['responseData']['nodes']), pool_size), start=1)]
 
     def req(self, req):
@@ -555,6 +531,9 @@ class ExaConnection(object):
         finally:
             self._req_lock.release()
 
+        if not recv_data:
+            raise ExaCommunicationError(self, "Empty WebSocket response, connection was likely closed")
+
         # Parse response
         ret = self.json_decode(recv_data)
         self.logger.debug_json(f'WebSocket response #{local_req_count}', ret)
@@ -629,7 +608,7 @@ class ExaConnection(object):
         self.login_time = time.time() - start_ts
 
         if self.options['compression']:
-            self._ws_send = lambda x: self._ws.send_binary(zlib.compress(x.encode(), 1))
+            self._ws_send = lambda x: self._ws.send_binary(zlib.compress(x.encode() if isinstance(x, str) else x, 1))
             self._ws_recv = lambda: zlib.decompress(self._ws.recv())
 
     def _login_password(self):
@@ -669,35 +648,43 @@ class ExaConnection(object):
         """
         Init websocket connection
         Connection redundancy is supported
-        Specific Exasol host is randomly selected for every connection attempt
+        Specific Exasol node is randomly selected for every connection attempt
         """
-        host_port_list = self._process_dsn(self.options['dsn'])
+        dsn_items = self._process_dsn(self.options['dsn'])
         failed_attempts = 0
 
         ws_prefix = 'wss://' if self.options['encryption'] else 'ws://'
         ws_options = self._get_ws_options()
 
-        for host, port in host_port_list:
-            self.logger.debug(f"Connection attempt [{host}:{port}]")
+        for hostname, ipaddr, port, fingerprint in dsn_items:
+            self.logger.debug(f"Connection attempt [{ipaddr}:{port}]")
+
+            # Use correct hostname matching IP address for each connection attempt
+            if self.options['encryption']:
+                ws_options['sslopt']['server_hostname'] = hostname
 
             try:
-                self._ws = websocket.create_connection(f'{ws_prefix}{host}:{port}', **ws_options)
+                self._ws = websocket.create_connection(f'{ws_prefix}{ipaddr}:{port}', **ws_options)
+            except Exception as e:
+                self.logger.debug(f'Failed to connect [{ipaddr}:{port}]: {e}')
+
+                failed_attempts += 1
+
+                if failed_attempts == len(dsn_items):
+                    raise ExaConnectionFailedError(self, 'Could not connect to Exasol: ' + str(e))
+            else:
                 self._ws.settimeout(self.options['socket_timeout'])
 
-                self.ws_host = host
+                self.ws_ipaddr = ipaddr
                 self.ws_port = port
 
                 self._ws_send = self._ws.send
                 self._ws_recv = self._ws.recv
 
+                if fingerprint:
+                    self._validate_fingerprint(fingerprint)
+
                 return
-            except Exception as e:
-                self.logger.debug(f'Failed to connect [{host}:{port}]: {e}')
-
-                failed_attempts += 1
-
-                if failed_attempts == len(host_port_list):
-                    raise ExaConnectionFailedError(self, 'Could not connect to Exasol: ' + str(e))
 
     def _get_ws_options(self):
         options = {
@@ -707,13 +694,14 @@ class ExaConnection(object):
         }
 
         if self.options['encryption']:
-            # Exasol JDBC / ODBC drivers do not check validity of certificates by default,
-            # so PyEXASOL follows this behaviour
-            #
-            # It is possible to enable cert validation, use custom certificate and set other SSL options by using
-            # "websocket_sslopt" argument, which is passed directly to WebSocket client constructor
+            # Use "websocket_sslopt" argument to set custom SSL options, which are passed directly to WebSocket client constructor
             if self.options['websocket_sslopt']:
                 options['sslopt'] = self.options['websocket_sslopt']
+            # OpenID tokens are normally used for Exasol SAAS
+            # Strict certificate verification is enabled by default
+            elif self.options['access_token'] or self.options['refresh_token']:
+                options['sslopt'] = {'cert_reqs': ssl.CERT_REQUIRED}
+            # Certification verification is disabled for other use cases (e.g. Docker container, "on-premises" Exasol setup)
             else:
                 options['sslopt'] = {'cert_reqs': ssl.CERT_NONE}
 
@@ -743,19 +731,23 @@ class ExaConnection(object):
 
     def _process_dsn(self, dsn):
         """
-        Parse DSN, expand ranges and resolve IP addresses for all hosts
-        Return list of host:port tuples in random order
-        Randomness is useful to guarantee good distribution of workload across all nodes
+        Parse DSN, expand ranges and resolve IP addresses for all hostnames
+        Return list of (hostname, ip_address, port) tuples in random order
+        Randomness is required to guarantee proper distribution of workload across all nodes
         """
         if len(dsn.strip()) == 0:
             raise ExaConnectionDsnError(self, 'Connection string is empty')
 
-        current_port = None
+        current_port = constant.DEFAULT_PORT
+        current_fingerprint = None
+
         result = []
 
-        dsn_re = re.compile(r'^(?P<host_prefix>.+?)'
+        dsn_re = re.compile(r'^(?P<hostname_prefix>.+?)'
                             # Optional range (e.g. myxasol1..4.com)
-                            r'(?:(?P<range_start>\d+)\.\.(?P<range_end>\d+)(?P<host_suffix>.*?))?'
+                            r'(?:(?P<range_start>\d+)\.\.(?P<range_end>\d+)(?P<hostname_suffix>.*?))?'
+                            # Optional fingerprint (e.g. myexasol1..4.com/135a1d2dce102de866f58267521f4232153545a075dc85f8f7596f57e588a181)
+                            r'(?:/(?P<fingerprint>[0-9A-Fa-f]+))?'
                             # Optional port (e.g. myexasol1..4.com:8564)
                             r'(?::(?P<port>\d+)?)?$'
                             )
@@ -770,13 +762,16 @@ class ExaConnection(object):
             if not m:
                 raise ExaConnectionDsnError(self, f'Could not parse connection string part [{part}]')
 
-            # Optional port was specified
+            # Optional port
             if m.group('port'):
                 current_port = int(m.group('port'))
 
-            # If current port is still empty, use default port
-            if current_port is None:
-                current_port = constant.DEFAULT_PORT
+            # Optional fingerprint
+            if m.group('fingerprint'):
+                current_fingerprint = m.group('fingerprint').upper()
+
+                if not self.options['encryption']:
+                    raise ExaConnectionDsnError(self, 'Fingerprint was specified in connection string, but encryption is not enabled')
 
             # Hostname or IP range was specified, expand it
             if m.group('range_start'):
@@ -788,28 +783,35 @@ class ExaConnection(object):
                 zfill_width = len(m.group('range_start'))
 
                 for i in range(int(m.group('range_start')), int(m.group('range_end')) + 1):
-                    host = f"{m.group('host_prefix')}{str(i).zfill(zfill_width)}{m.group('host_suffix')}"
-                    result.extend(self._resolve_host(host, current_port))
+                    hostname = f"{m.group('hostname_prefix')}{str(i).zfill(zfill_width)}{m.group('hostname_suffix')}"
+                    result.extend(self._resolve_hostname(hostname, current_port, current_fingerprint))
             # Just a single hostname or single IP address
             else:
-                result.extend(self._resolve_host(m.group('host_prefix'), current_port))
+                result.extend(self._resolve_hostname(m.group('hostname_prefix'), current_port, current_fingerprint))
 
         random.shuffle(result)
 
         return result
 
-    def _resolve_host(self, host, port):
+    def _resolve_hostname(self, hostname, port, fingerprint):
         """
-        Resolve all IP addresses for host and add port
-        It also checks that all hosts mentioned in DSN can be resolved
+        Resolve all IP addresses for hostname and add port
+        It also implicitly checks that all hostnames mentioned in DSN can be resolved
         """
         try:
-            hostname, aliaslist, ipaddrlist = socket.gethostbyname_ex(host)
+            hostname, alias_list, ipaddr_list = socket.gethostbyname_ex(hostname)
         except OSError:
-            raise ExaConnectionDsnError(self, f'Could not resolve IP address of host [{host}] '
+            raise ExaConnectionDsnError(self, f'Could not resolve IP address of hostname [{hostname}] '
                                               f'derived from connection string')
 
-        return [(ipaddr, port) for ipaddr in ipaddrlist]
+        return [(hostname, ipaddr, port, fingerprint) for ipaddr in ipaddr_list]
+
+    def _validate_fingerprint(self, provided_fingerprint):
+        server_fingerprint = hashlib.sha256(self._ws.sock.getpeercert(True)).hexdigest().upper()
+
+        if provided_fingerprint != server_fingerprint:
+            raise ExaConnectionFailedError(self, f"Provided fingerprint [{provided_fingerprint}] did not match "
+                                                 f"server fingerprint [{server_fingerprint}]")
 
     def _init_logger(self):
         self.logger = self.cls_logger(self, constant.DRIVER_NAME)
@@ -820,26 +822,29 @@ class ExaConnection(object):
         self.format = self.cls_formatter(self)
 
     def _init_json(self):
-        # rapidjson is well maintained library with acceptable performance, good choice
         if self.options['json_lib'] == 'rapidjson':
             import rapidjson
 
-            self.json_encode = lambda x: rapidjson.dumps(x, number_mode=rapidjson.NM_NATIVE)
+            self.json_encode = lambda x, indent=False: rapidjson.dumps(x, number_mode=rapidjson.NM_NATIVE, indent=2 if indent else None, ensure_ascii=False)
             self.json_decode = lambda x: rapidjson.loads(x, number_mode=rapidjson.NM_NATIVE)
 
-        # ujson provides best performance in our tests, but it is abandoned by maintainers
         elif self.options['json_lib'] == 'ujson':
             import ujson
 
-            self.json_encode = ujson.dumps
-            self.json_decode = ujson.loads
+            self.json_encode = lambda x, indent=False: ujson.dumps(x, indent=2 if indent else 0, ensure_ascii=False)
+            self.json_decode = lambda x: ujson.loads(x)
+
+        elif self.options['json_lib'] == 'orjson':
+            import orjson
+
+            self.json_encode = lambda x, indent=False: orjson.dumps(x, option=orjson.OPT_INDENT_2 | orjson.OPT_APPEND_NEWLINE if indent else 0)
+            self.json_decode = lambda x: orjson.loads(x)
 
-        # json from Python stdlib, very safe choice, but slow
         elif self.options['json_lib'] == 'json':
             import json
 
-            self.json_encode = json.dumps
-            self.json_decode = json.loads
+            self.json_encode = lambda x, indent=False: json.dumps(x, indent=2 if indent else None, ensure_ascii=False)
+            self.json_decode = lambda x: json.loads(x)
 
         else:
             raise ValueError(f"Unsupported json library [{self.options['json_lib']}]")
